/*
 * Copyright 2008 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.lucis.core.impl;

import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.concurrent.atomic.AtomicLong;

import net.sf.lucis.core.Checkpoints;
import net.sf.lucis.core.ReindexingStore;

import org.apache.lucene.analysis.standard.StandardAnalyzer;
import org.apache.lucene.index.IndexWriter;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.FSDirectory;

import com.google.common.base.Preconditions;

/**
 * File system-based reindexing store.
 * @author Andres Rodriguez
 */
public class ReindexingFSStore implements ReindexingStore {
	/** Name of the control file. */
	private static final String STATUS_FILE = "status.ctl";
	/** Relative path to first copy. */
	private static final String COPY01 = "copy01";
	/** Relative path to second copy. */
	private static final String COPY02 = "copy02";
	/** Main directory. */
	private final File file;
	/** Status file. */
	private final File statusFile;

	/** Status. */
	private enum Status {
		NULL, CREATING, DONE01, DONE02
	}

	private Status status = Status.NULL;
	/** First copy. */
	private final Copy copy01;
	/** Second copy. */
	private final Copy copy02;
	/** Version. */
	private final AtomicLong version = new AtomicLong(0);

	private static Status readStatus(File file) {
		if (!file.exists()) {
			return Status.NULL;
		}
		try {
			final FileInputStream stream = new FileInputStream(file);
			try {
				final String str = Checkpoints.ofStringLine().read(stream);
				if (str == null || str.length() == 0) {
					return Status.NULL;
				}
				try {
					return Enum.valueOf(Status.class, str);
				} catch (IllegalArgumentException e) {
					return Status.NULL;
				}
			} finally {
				close(stream);
			}
		} catch (FileNotFoundException fnfe) {
			// Nothing.
		} catch (IOException e) {
			throw new IllegalStateException(e);
		}
		return Status.NULL;
	}

	private static void close(Closeable closeable) {
		if (closeable == null) {
			return;
		}
		try {
			closeable.close();
		} catch (IOException ioe) {
		}
	}

	public ReindexingFSStore(final String indexDir) {
		try {
			this.file = new File(indexDir);
			Preconditions.checkArgument(file.exists() && file.isDirectory(), "Invalid index directory");
			this.statusFile = new File(file, STATUS_FILE);
			Preconditions.checkArgument(!statusFile.isDirectory(), "Status file cannot be a directory");
			this.copy01 = new Copy(file, COPY01);
			this.copy02 = new Copy(file, COPY02);
			this.status = readStatus(this.statusFile);
			if (Status.NULL.equals(status)) {
				IndexWriter w = new IndexWriter(copy01.directory, new StandardAnalyzer(), true, IndexWriter.MaxFieldLength.UNLIMITED);
				w.commit();
				w.close();
				reindexed();
			}
		} catch (IOException e) {
			throw new IllegalStateException(e);
		}
	}

	public Directory getDirectory() {
		switch (status) {
		case NULL:
			return null;
		case CREATING:
			return copy01.directory;
		case DONE01:
			return copy01.directory;
		case DONE02:
			return copy02.directory;
		}
		throw new AssertionError();
	}

	public Directory getDestinationDirectory() {
		switch (status) {
		case DONE01:
			return copy02.directory;
		default:
			return copy01.directory;
		}
	}

	public void reindexed() {
		switch (status) {
		case DONE01:
			status = Status.DONE02;
			break;
		default:
			status = Status.DONE01;
			break;
		}
		version.incrementAndGet();
		writeStatus();
	}

	private void writeStatus() {
		try {
			final FileOutputStream fos = new FileOutputStream(statusFile, false);
			try {
				Checkpoints.ofStringLine().write(status.name(), fos);
			} finally {
				close(fos);
			}
		} catch (IOException e) {
		}
	}

	public Object getVersion() {
		return version.get();
	}

	private static final class Copy {
		private final File file;
		private final Directory directory;

		Copy(File base, String path) throws IOException {
			this.file = new File(base, path);
			
			Preconditions.checkArgument(!file.exists() || (file.exists() && file.isDirectory()), "Invalid index copy directory");
			if (!file.exists()) {
				file.mkdir();
			}
			this.directory = FSDirectory.getDirectory(this.file);
		}

	}

}
