/*
 * Copyright 2008 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.lucis.core.impl;

import java.util.concurrent.Callable;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;

import net.sf.derquinsej.concurrent.ActiveObjectStatus;
import net.sf.derquinsej.concurrent.ActiveObjectSupport;
import net.sf.lucis.core.Delays;
import net.sf.lucis.core.FullIndexer;
import net.sf.lucis.core.IndexException;
import net.sf.lucis.core.IndexStatus;
import net.sf.lucis.core.IndexerService;
import net.sf.lucis.core.ReindexingStore;
import net.sf.lucis.core.ReindexingWriter;

/**
 * Default reindexing service.
 * @author Andres Rodriguez
 */
public class ReindexingIndexerService implements IndexerService {
	/** Default name. */
	private static final String NAME = "ReindexingIndexerService";
	/** Writer. */
	private final ReindexingWriter writer;
	/** Index status. */
	private volatile IndexStatus status = IndexStatus.OK;
	/** Index store. */
	private final ReindexingStore store;
	/** Indexer. */
	private final FullIndexer indexer;
	/** Active object support. */
	private final ActiveObjectSupport support;
	/** Executor. */
	private volatile ScheduledExecutorService executor;
	/** Delays. */
	private volatile Delays delays = Delays.constant(600000);
	/** Log. */
	private Logger log = Logger.getLogger(getClass().getName());
	/** Service name. */
	private String name = NAME;

	public ReindexingIndexerService(ReindexingStore store, ReindexingWriter writer, FullIndexer indexer) {
		this.store = store;
		this.writer = writer;
		this.indexer = indexer;
		final Runnable onStart = new Runnable() {
			public void run() {
				executor = Executors.newSingleThreadScheduledExecutor();
				schedule(0L);
			}
		};
		final Runnable onStop = new Runnable() {
			public void run() {
				final ScheduledExecutorService ses = executor;
				if (ses != null) {
					executor = null;
					ses.shutdown();
					try {
						while (!ses.isTerminated()) {
							ses.awaitTermination(5L, TimeUnit.SECONDS);
						}
					} catch (InterruptedException e) {
					}
				}
			}
		};
		this.support = new ActiveObjectSupport(onStart, null, onStop);
	}

	/* CONFIGURABLE PROPERTIES */

	public void setLogName(final String name) {
		this.log = Logger.getLogger(name);
	}

	public void setDelays(Delays delays) {
		this.delays = delays;
	}

	public void setName(String name) {
		this.name = (name != null && name.length() > 0) ? name : NAME;
	}
	
	/* END CONFIGURABLE PROPERTIES. */

	/* PUBLIC API */

	public IndexStatus getIndexStatus() {
		return status;
	}

	public ActiveObjectStatus getStatus() {
		return support.getStatus();
	}

	public void start() {
		support.start();
	}

	public void stop() {
		support.stop();
	}

	/* END PUBLIC API */

	private void schedule(long delay) {
		final ScheduledExecutorService ses = executor;
		if (ses != null) {
			ses.schedule(new Task(), delay, TimeUnit.MILLISECONDS);
		}
	}

	private void schedule(IndexStatus status, long delay) {
		this.status = status;
		schedule(delay);
	}
	
	private String format(String s) {
		return String.format("Service [%s]: %s", name, s);
	}

	private final class Task implements Runnable {
		public void run() {
			final Callable<Object> callable = new Callable<Object>() {
				public Object call() throws Exception {
					writer.reindex(store, indexer);
					return null;
				}
			};
			try {
				MayFail.run(callable);
				log.finest(format("Reindexing complete."));
				schedule(IndexStatus.OK, delays.getNormal());
			} catch (IndexException e) {
				log.log(Level.SEVERE, format("Index exception while reindexing"), e);
				schedule(e.getStatus(), delays.getError());
			} catch (Exception ex) {
				log.log(Level.SEVERE, format("Exception while reindexing"), ex);
				schedule(IndexStatus.ERROR, delays.getError());
			}
		}
	}

}
